# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
import chevron
import pathlib

from .interfaces import Destination, DestinationDecorator


class FetchFileGenerator(DestinationDecorator):
    """
    Generates a shell script which can be used to fetch all the files related to a request.
    The script uses "wget".
    """

    def __init__(self, underlying: Destination):
        super().__init__(underlying)

        # generate the file
        self.fetch_script = underlying.create_file("fetch-all.sh")

        # write some lines to it
        self.fetch_script.file().writelines([b"#!/bin/sh\n", b"\n"])

    def add_file(self, file: pathlib.Path, relative_path: str):
        super().add_file(file, relative_path)

        url = self.underlying.results()["url"]

        # add a line to the fetch script for this file
        self.fetch_script.file().write(
            chevron.render('wget -r --reject "index.html*" -np -nH --cut-dirs=3 {{ url }}/{{ relative_path }}\n', {
                'url': url,
                'relative_path': relative_path
            }).encode("utf8")
        )

    def close(self):
        # first close the script
        self.fetch_script.close()

        # proceed
        super().close()

    def __str__(self):
        return str(self.underlying) + " with fetch-all.sh"
